//
//  Manitest.cpp
//  MANITEST
//
//  Created by Alhussein Fawzi on 20.05.15.
//  Copyright (c) 2015 Alhussein Fawzi. All rights reserved.
//

#include "Manitest.h"
#include "Stencil.h"

// Constructor
Manitest::Manitest(cv::Mat _img, Transformation* _trans, Stencil* _stencil, Classifiers* _classifiers) {
    img = _img;
    trans = _trans;
    stencil = _stencil; stencil->parentClass = this;
    classifiers = _classifiers;
    p = _trans->p;
    N = 50000;
}

void Manitest::initialize() {
    // Initialize elements
    dist.push_back(0);
    id2Coords.push_back(std::vector<int>(p,0)); // Origin has ID=0, and all coordinates=0
    coords2Id.insert(std::pair<std::vector<int> ,int>(std::vector<int>(p,0),0));
    visited.push_back(false);
    metrics.push_back(cv::Mat(0,0,CV_32F));
    
    // Populate the label set, if the user specified a set of classifiers
    if (classifiers != nullptr) {
        labelsOrig = classifiers->classify(img);
    }
    trans->img = &img;
}

void Manitest::run() {
    
    initialize();
    int i, j, j_, k_, c;
    cv::Mat X, v;
    cv::Mat invM, ones;
    double oo, ov, vv;
    std::vector<bool> mask;
    int ncl = 0;
    
    if (classifiers != nullptr) ncl = (int) classifiers->ListClassifiers.size();
    
    mask = std::vector<bool>(ncl, true);
    scores = std::vector<double>(ncl);
    params = std::vector<std::vector<double> >(ncl);
    images = std::vector<cv::Mat>(ncl);
    
    double min_el;
    
    for (i=0; i<N; i++) {
        
        if (i%100 == 0) {
            std::cout << "i = " << i << std::endl;
        }
        
        int id = -1; min_el = INFINITY;
        
        // Determine the minimal element in dist that is unvisited
        for (j=0; j<dist.size(); j++) {
            if (visited[j] == false && dist[j] < min_el) {
                min_el = dist[j];
                id = j;
            }
        }
        if (id == -1) {
            std::cout << "I cannot continue anymore. Breaking." << std::endl;
            return;
        }
        
        visited[id] = true;
        std::vector<double> para_accepted = stencil->coords2Para(&id2Coords[id]);
        
        // Check whether the current transformed image is wrongly classified
        if (classifiers != nullptr) {
            cv::Mat trim = trans->transformImage(para_accepted);
            std::vector<int> res = classifiers->classify(trim, mask);
            for (c=0; c<res.size(); c++) {
                if (!mask[c]) continue;
                
                if (res[c] != labelsOrig[c]) {
                    scores[c] = dist[id] / norm(img);
                    params[c] = para_accepted;
                    images[c] = trim;
                    
                    std::cout << ">>> Classifier #" << c << ": (";
                    for (int l=0; l<para_accepted.size(); l++) {
                        std::cout << para_accepted[l];
                        
                        if (l < para_accepted.size()-1) {
                            std::cout << ", ";
                        }
                    }
                    std::cout << "), iter #" << i << ", score = " << scores[c] << std::endl;
                    
                    mask[c] = false;
                    
                    // Show image 
                    /* stringstream sstm; sstm << "Classifier #" << c;
                    showImage(sstm.str(), trim); */
                }
            }
            
            // Check whether all classifiers are fooled?
            // if (std::all_of(std::begin(mask), std::end(mask), [](bool i) { return !i; })) {
            if (checkAllFalse(mask)) {
                // All numbers are false
                std::cout << "I fooled all classifiers... Quitting.\n" << std::endl;
                return;
            }
        }
        
        std::vector<int> ids_rStencil = stencil->getReversedStencilsId(&id2Coords[id]);
        
        for (j=0; j<ids_rStencil.size(); j++) {
            
            
            int id_curPoint = ids_rStencil[j];
            std::vector<double> curPara = stencil->coords2Para(&id2Coords[id_curPoint]);
            
            if (visited[id_curPoint] == true) {
                // Already visited, nothing to do
                continue;
            }
            
            // Generate the metric at id_neighbour (if not already created)
            if (metrics[id_curPoint].empty()) {
                metrics[id_curPoint] = stencil->computeMetric(curPara);
            }
            
            dist[id_curPoint] = std::min(dist[id_curPoint], dist[id] + sqrt(norm_sq_mat(metrics[id_curPoint], cv::Mat(curPara) - cv::Mat(para_accepted))));
                
            
            stencil->genSimplicesId(&id2Coords[id_curPoint], &id2Coords[id]);
            
            std::vector<std::vector<int> > simplexSetId = stencil->simplexSetId;
            
            for (j_ = 0; j_<simplexSetId.size(); j_++) {
                
                cv::Mat X((int) simplexSetId[j_].size(), p, CV_64F); // Double precision
                cv::Mat v((int) simplexSetId[j_].size(), 1, CV_64F); // Double precision
                
                for (k_ = 0; k_ < simplexSetId[j_].size(); k_++) {
                    
                    // point k_ in simplex j_
                    std::vector<double> para = stencil->coords2Para(&id2Coords[simplexSetId[j_][k_]]);
                    subtractVec(&para, &curPara);
                    X.row(k_) = cv::Mat(para).t();
                    v.row(k_) = dist[simplexSetId[j_][k_]];
                }
                
                invM = (X * metrics[id_curPoint] * X.t()).inv();
                ones = cv::Mat::ones((int) simplexSetId[j_].size(), 1, CV_64F);
                
                oo = norm_sq_mat(invM, ones);
                ov = inner_product_mat(invM, ones, v);
                vv = norm_sq_mat(invM, v);
                
                double Delta = ov * ov - oo * (vv - 1);
                if (Delta < 0) continue;
                double cand_d = (ov + sqrt(Delta)) / oo;
                if (cand_d >= dist[id_curPoint]) continue;
                
                // Make sure it is inside the simplex
                cv::Mat w = invM * (cand_d - v); double minval;
                minMaxLoc(w, &minval);
                if (minval <= 0) continue;
                
                dist[id_curPoint] = cand_d;
            }
            
            
        }
    }
    
}


cv::Mat Manitest::constructDistanceMap() {
    
    
    int m1, m2, M1, M2;
    m1 = INFINITY; m2 = INFINITY;
    M1 = -INFINITY; M2 = -INFINITY;
    
    int i;
    for (i=0; i<id2Coords.size(); i++) {
        if (id2Coords[i][0] < m1) {
            m1 = id2Coords[i][0];
        }
        if (id2Coords[i][0] > M1) {
            M1 = id2Coords[i][0];
        }
        if (id2Coords[i][1] < m2) {
            m2 = id2Coords[i][1];
        }
        if (id2Coords[i][1] > M2) {
            M2 = id2Coords[i][1];
        }
    }
    
    cv::Mat distanceMap((M1 - m1) + 1, (M2 - m2) + 1, CV_64F);
    
    for (i=0; i<id2Coords.size(); i++) {
        if (!visited[i]) continue;
        point cur_coord = id2Coords[i];
        distanceMap.at<double>(cur_coord[0] - m1, cur_coord[1] - m2) = dist[i];
    }
    
    showImage("Distance map", distanceMap);
    return distanceMap;
    
}